//
// GABitStringGenome.cpp - GA bit string genome
// representation, operators
/////////////////////////////////////////////
#include "stdafx.h"
#include "gagenome.h"
#include "assert.h"
#include "gasolver.h"
//-----------------------------------------
GABitStringGenome::GABitStringGenome(int size)	  
{
	m_bits = NULL;
	m_size=0;
	SetSize(size);
	Clear();
};

//-----------------------------------------
GABitStringGenome::~GABitStringGenome() 
{
	if (m_bits)
 	  delete [] m_bits;
	m_bits = NULL;
};


//-----------------------------------------
void GABitStringGenome::SetSize( int size)
{
	if (m_size != size)
	{
		m_size = size;
		if (m_bits)
			delete [] m_bits;
		if (m_size != 0)
		{
			m_bits = new BYTE [(m_size + 7)/8];
		}
		else
		{
			m_bits = NULL;
		}
		Clear();
	}
}

//-----------------------------------------
GABitStringGenome& GABitStringGenome::operator 
   = (const GABitStringGenome& p)
{
	m_obj = p.m_obj;
	m_fitness = p.m_fitness;
	m_selProb = p.m_selProb;
	m_selected = p.m_selected;
	SetSize( p.m_size);
	memcpy( m_bits, p.m_bits, (m_size +7)/8);
	return *this;
}

//-----------------------------------------
void GABitStringGenome::Set()
{
	if (m_bits)
		memset(m_bits, 0xFF,(m_size+7)/8);
};

//-----------------------------------------
void GABitStringGenome::Set(int i)
{
	assert(m_bits != 0);
	m_bits[i/8]  |= (BYTE) (0x01<<(i%8));
};

//-----------------------------------------
void GABitStringGenome::Set( int fromBit, int toBit)
{
	//
	// Find the specified bits in the starting and ending BYTEs.
	//
	int startByte  = fromBit/8;
	int startBit	= fromBit % 8;
	BYTE	startMask = 0xFF;
	startMask <<= startBit;

	int endByte  = toBit/8;
	int endBit	= toBit % 8;
	BYTE	endMask  = 0xFF;
	endMask >>= (7-endBit);
	//
	// Set these bits in the starting and ending BYTEs.
	//
	if (startByte == endByte)
	{
		m_bits[startByte] |= (startMask&endMask);
	}
	else
	{
		m_bits[startByte] |= startMask;
		m_bits[endByte]	|= endMask;
	}
	//
	//  Now set everything in between.
	//
	int size = endByte - startByte - 1;
	if (size > 0)
	{
		memset( m_bits + startByte + 1, 0xFF, size);
	}
}

//-----------------------------------------
void GABitStringGenome::Clear()
{
	if (m_bits)
		memset(m_bits, 0, (m_size+7)/8);
};

//-----------------------------------------
void GABitStringGenome::Clear(int bit)
{
	assert(m_bits != 0);
	m_bits[bit/8]  &= (BYTE) ~((BYTE)0x01<<(bit%8));
};



//-----------------------------------------
void GABitStringGenome::Clear( int start, int len)
{
	//
	// Find the specified bits in the starting and ending BYTEs.
	//
	int startByte  = start/8;
	int startBit	= start % 8;
	BYTE	startMask = 0xFF;
	startMask <<= startBit;

	int toBit = start + len - 1;
	int endByte  = toBit/8;
	int endBit	= toBit % 8;
	BYTE	endMask  = 0xFF;
	endMask >>= (7-endBit);
	//
	//  Clear the bits in the starting and ending BYTEs.
	//
	if (startByte == endByte)
	{
		m_bits[startByte] &= (BYTE)~(startMask&endMask);
	}
	else
	{
		m_bits[startByte] &= (BYTE)~startMask;
		m_bits[endByte] &= (BYTE)~endMask;
	}

	//
	//  Now clear everything in between.
	//
	int size = endByte - startByte - 1;
	if (size > 0)
	{
		memset( m_bits + startByte + 1, 0, size);
	}
}


//-----------------------------------------
bool GABitStringGenome::IsSet(int bit)
{
	assert( m_bits != 0);
	if (m_bits[bit/8]&(BYTE)(0x01<<(bit%8)))
		return true;
	else
		return false;
}

//-----------------------------------------
void GABitStringGenome::Randomize()
{
	BYTE *byt = m_bits;
	int cnt = m_size/8;
	while (cnt-- > 0)
	{
		*byt = (BYTE)RandomNum(0x100);
		byt++;
	}
	
}


//-----------------------------------------
void GABitStringGenome::CrossOver( GAGenome *gen)
{
	GABitStringGenome *genome = (GABitStringGenome *)gen;
	// Generate a crossover point at random.
	int xover = RandomNum(m_size);
	//
	// determine the crossover byte.
	int xoverByte = xover/8;
	int xoverBit = xover % 8;
	if (xoverBit > 0)
		xoverByte++;
	int nBytes = (m_size + 7)/8;
	for (int i=xoverByte;i<nBytes;i++)
	{
		BYTE a = genome->m_bits[i];
		BYTE b = m_bits[i];
		m_bits[i] = a;
		genome->m_bits[i] = b;
	}
	//
	// now swap bits within crossover byte.
	if (xoverBit > 0)
	{
		BYTE	mask2 = 0xFF;
		mask2 <<=  xoverBit;
		BYTE mask1 = ~mask2;
		xoverByte--;
		BYTE a = (mask1&m_bits[xoverByte]) | (mask2&genome->m_bits[xoverByte]);
		BYTE b = (mask1&genome->m_bits[xoverByte]) | (mask2&m_bits[xoverByte]);
		m_bits[xoverByte] = a;
		genome->m_bits[xoverByte] = b;
	}
}


//-----------------------------------------
void GABitStringGenome::Mutate( double pmutate)
{
	// for each bit
	for (int k=0;k<m_size;k++)
	{
		// generate a random mumber
		double pk = RandomNum(10000)/10000.0;
		// flip the bit if less than mutation probablity
		if (pk < pmutate)
		{
			if (IsSet( k))
				Clear(k);
			else
				Set(k);
		}
	}
}


