//
//  knapsack.cpp
#include "stdafx.h"
#include "knapsack.h"
#include "math.h"

int RandomNum(int maxValPlusOne)
{
   int r = rand();
   int rnum = (r*maxValPlusOne)/(RAND_MAX+1);
   return rnum;
}

CKnapSackProblem::CKnapSackProblem()
{
	m_numItems = 40;
	m_maxWeight = 2000;
	m_avgWeightPerItem = 100;
	m_weightVariance = 50;
	m_avgProfitPerItem = 100;
	m_profitVariance = 50;
	m_item = NULL;
	Regenerate();
}

CKnapSackProblem::~CKnapSackProblem()
{
	if (m_item != NULL)
		delete [] m_item;
}
//
//  Using the parameters provided.
//  Generate a new knapsack problem
//  with m_numItems items with various
//  weights and profits.
void CKnapSackProblem::Regenerate()
{
	srand(17);
	m_item = NULL;
	if (m_numItems > 0)
		m_item = new CKnapSackItem [m_numItems];
	for (int i=0;i<m_numItems;i++)
	{
		double lo = m_avgWeightPerItem*(1.0 - m_weightVariance/100.0/2.0);
		if (lo <= 1.0)
			lo = 1.0;
		double hi = m_avgWeightPerItem*(1.0 + m_weightVariance/100.0/2.0);
		double result = lo + (hi-lo)*RandomNum(10000)/10000.0;
		result = floor(result);
		m_item[i].m_weight = result;

		lo = m_avgProfitPerItem*(1.0 - m_profitVariance/100.0/2.0);
		if (lo <= 1.0)
			lo = 1.0;
		hi = m_avgProfitPerItem*(1.0 + m_profitVariance/100.0/2.0);
		result = lo + (hi-lo)*RandomNum(10000)/10000.0;
		result = floor(result);
		m_item[i].m_profit = result;
		m_item[i].m_isIncluded = false;
	}
}

void  CKnapSackProblem::GetTotals( int &totalItems, double &totalProfit, double &totalWeight)
{
	int nItems =0;
	double profit =0;
	double weight =0;
	for (int i=0;i<m_numItems;i++)
	{
		if (m_item[i].m_isIncluded)
		{
			weight += m_item[i].m_weight;
			profit += m_item[i].m_profit;
			nItems++;
		}
	}
	totalItems = nItems;
	totalProfit = profit;
	totalWeight = weight;
}
