// KnapSackGASolver.cpp : Implementation of CKnapSackGASolver
#include "stdafx.h"
#include "Ksgasolve.h"
#include "KnapSackGASolver.h"
#include "ksgenome.h"
#include "gasolver.h"

/////////////////////////////////////////////////////////////////////////////
// CKnapSackGASolver
int      KSGenome::m_maxItems = 0;
double*  KSGenome::m_profit   = NULL;
double*  KSGenome::m_weight   = NULL;
double   KSGenome::m_maxCapacity=0;
double   KSGenome::m_penalty  = 0 ;


STDMETHODIMP CKnapSackGASolver::Solve(int *nItems, double *profits, double *weights, double maxWeight, BOOL **isIncluded, double *totalProfit)
{
   KSGenome bestSoln;
   if (bestSoln.m_profit)
      delete [] bestSoln.m_profit;
   if (bestSoln.m_weight)
      delete [] bestSoln.m_weight;
   int maxItems = *nItems;
   bestSoln.m_maxItems = maxItems;
   bestSoln.m_maxCapacity = maxWeight;
   bestSoln.m_profit = new double[maxItems];
   bestSoln.m_weight = new double[maxItems];

   double maxProfitRatio = 0;
   for (int i=0;i<*nItems;i++)
   {
      if (weights[i] > 0)
      {
         double profitRatio = profits[i]/weights[i];
         if (profitRatio > maxProfitRatio)
            maxProfitRatio = profitRatio;
      }
      bestSoln.m_profit[i] = profits[i];
      bestSoln.m_weight[i] = weights[i];
   }
   bestSoln.m_penalty = 3*maxProfitRatio;
   //
   //  Weights, profits and penalties are set.
   //  Now solve using GASolver template class.
   GASolver <KSGenome> gaSolver;
   gaSolver.SetPopSize(200);
   gaSolver.SetNumGenerations(1000);
   gaSolver.SetMutationFactor(0.01);
   gaSolver.SetCrossOverFactor(0.40);
   gaSolver.SetFitnessSigma(1.5);
   gaSolver.Maximize(bestSoln);
   *totalProfit = bestSoln.Evaluate();
   *isIncluded = (BOOL *)CoTaskMemAlloc(maxItems*sizeof(BOOL));
   for (i=0;i<maxItems;i++)
   {
      if (bestSoln.IsSet(i))
      {
         (*isIncluded)[i] = TRUE;
      }
      else
      {
         (*isIncluded)[i] = FALSE;
      }
   }
   if (bestSoln.m_profit)
      delete [] bestSoln.m_profit;
   if (bestSoln.m_weight)
      delete [] bestSoln.m_weight;
   bestSoln.m_maxItems = 0;
   bestSoln.m_profit   = NULL;
   bestSoln.m_weight   = NULL;
   return S_OK;
}
