#if !defined ( THETA_H_INCLUDED )
#define THETA_H_INCLUDED

// This algorithm comes from an article in "Electronics," a
// trade journal. It was published in an article on a new
// chip in 1980. The original publication had several 
// significant errors in the algorithm, but I was able to 
// fix them. They gave the impression of being intentionally
// introduced. The chip was supposed to make database lookups
// extremely fast by doing most of the work in firmware.

#include <valarray>
//#include <algorithm>  // this is supposed to contain max, but it didn't on MS VC++ 6.0

using namespace std;

//===============================================================================================
template < typename T_a, typename T_b >  // std::max is not in the library that I have
T_a max ( const T_a& a, const T_b& b )
{
   return ( a >= b ? a : static_cast<T_a> ( b ) );
};

class CThetaSizeError
{
};

template < typename T >
class CThetaMatch
{
public:

//===============================================================================================
   CThetaMatch ( )
   {
      m_lMaxCharacter = 256;
      m_wf.resize( m_lMaxCharacter, 1.0 );
      m_wr.resize( m_lMaxCharacter, 0.8 );
      m_wc.resize( m_lMaxCharacter, 0.0 );
      m_bDoCompensation = false;
   };

//===============================================================================================
   CThetaMatch ( const size_t lMaxCharacter )
   {
      m_lMaxCharacter = lMaxCharacter;
      m_wf.resize( m_lMaxCharacter, 1.0 );
      m_wr.resize( m_lMaxCharacter, 0.8 );
      m_wc.resize( m_lMaxCharacter, 0.0 );
      m_bDoCompensation = false;
   };

//===============================================================================================
   CThetaMatch ( const valarray<T> vForwardWeights, const valarray<T> vReverseWeights )
   {
      m_lMaxCharacter = vForwardWeights.size( );
      m_wf = vForwardWeights;
      m_wr = vReverseWeights;
      m_wc.resize( m_lMaxCharacter, 0.0 );
      m_bDoCompensation = false;
   };

//===============================================================================================
   double operator ( ) ( const T& s, const T& t ) const
   {
      if ( s == t )
         return ( 0.0 );
      else
         return ( theta( s, t ) );
   };

//===============================================================================================
   void set_ThetaCompensation ( const valarray<double>& wc )
   {
      if ( wc.size( ) != m_lMaxCharacter ) throw CThetaSizeError( );
      m_wc = wc;
      m_bDoCompensation = true;
   };

//===============================================================================================
   void set_ThetaCompensation ( const double& dwc )
   {
      m_bDoCompensation = ( dwc > 0.0 );
      if ( m_bDoCompensation ) m_wc = dwc;
   };

//===============================================================================================
   void SetThetaWeights ( const valarray<double>& wf, const valarray<double>& wr )
   {
      if ( wf.size( ) != m_lMaxCharacter ) throw CThetaSizeError( );
      if ( wr.size( ) != m_lMaxCharacter ) throw CThetaSizeError( );
      m_wf = wf;
      m_wr = wr;
   };

//===============================================================================================
   void SetThetaWeights ( const double& wf, const double& wr )
   {
      m_wf = wf;
      m_wr = wr;
   };

//===============================================================================================
   double theta (const T& s, const T& t ) const
   {
      const double dTemp = totalM( s ) + totalM( t );
      if ( dTemp == 0.0 )
         return ( 1.0 );
      else
         return ( 1.0-(( m(s, t, 1, m_wf ) + m( s, t, -1, m_wr ) ) / dTemp ));
   };

//===============================================================================================
void m_fnSetEnglishFrequencies( )
{
/*
English letter frequencies
e 11.42
a 8.56
i 7.94
r 7.51
t 7.46
o 7.12
n 6.41
s 5.55
l 5.52
c 4.74
u 3.66
p 3.27
m 3.22
d 3.13
h 2.76
g 2.30
b 2.12
y 2.00
f 1.47
v 1.07
w 0.94
k 0.84
x 0.35
z 0.24
q 0.23
j 0.15

*/
   int i;
   for ( i=0; i<256; ++i )
   {
      wf[i] = 2.0;
   }
   m_wf[static_cast<int> ( 'e' ) ] = 11.42;
   m_wf[static_cast<int> ( 'a' ) ] = 8.56;
   m_wf[static_cast<int> ( 'i' ) ] = 7.94;
   m_wf[static_cast<int> ( 'r' ) ] = 7.51;
   m_wf[static_cast<int> ( 't' ) ] = 7.46;
   m_wf[static_cast<int> ( 'o' ) ] = 7.12;
   m_wf[static_cast<int> ( 'n' ) ] = 6.41;
   m_wf[static_cast<int> ( 's' ) ] = 5.55;
   m_wf[static_cast<int> ( 'l' ) ] = 5.52;
   m_wf[static_cast<int> ( 'c' ) ] = 4.74;
   m_wf[static_cast<int> ( 'u' ) ] = 3.66;
   m_wf[static_cast<int> ( 'p' ) ] = 3.27;
   m_wf[static_cast<int> ( 'm' ) ] = 3.22;
   m_wf[static_cast<int> ( 'd' ) ] = 3.13;
   m_wf[static_cast<int> ( 'h' ) ] = 2.76;
   m_wf[static_cast<int> ( 'g' ) ] = 2.30;
   m_wf[static_cast<int> ( 'b' ) ] = 2.12;
   m_wf[static_cast<int> ( 'y' ) ] = 2.00;
   m_wf[static_cast<int> ( 'f' ) ] = 1.47;
   m_wf[static_cast<int> ( 'v' ) ] = 1.07;
   m_wf[static_cast<int> ( 'w' ) ] = 0.94;
   m_wf[static_cast<int> ( 'k' ) ] = 0.84;
   m_wf[static_cast<int> ( 'x' ) ] = 0.35;
   m_wf[static_cast<int> ( 'z' ) ] = 0.24;
   m_wf[static_cast<int> ( 'q' ) ] = 0.23;
   m_wf[static_cast<int> ( 'j' ) ] = 0.15;

   m_wf[static_cast<int> ( 'E' ) ] = 11.42;
   m_wf[static_cast<int> ( 'A' ) ] = 8.56;
   m_wf[static_cast<int> ( 'I' ) ] = 7.94;
   m_wf[static_cast<int> ( 'R' ) ] = 7.51;
   m_wf[static_cast<int> ( 'T' ) ] = 7.46;
   m_wf[static_cast<int> ( 'O' ) ] = 7.12;
   m_wf[static_cast<int> ( 'N' ) ] = 6.41;
   m_wf[static_cast<int> ( 'S' ) ] = 5.55;
   m_wf[static_cast<int> ( 'L' ) ] = 5.52;
   m_wf[static_cast<int> ( 'C' ) ] = 4.74;
   m_wf[static_cast<int> ( 'U' ) ] = 3.66;
   m_wf[static_cast<int> ( 'P' ) ] = 3.27;
   m_wf[static_cast<int> ( 'M' ) ] = 3.22;
   m_wf[static_cast<int> ( 'D' ) ] = 3.13;
   m_wf[static_cast<int> ( 'H' ) ] = 2.76;
   m_wf[static_cast<int> ( 'G' ) ] = 2.30;
   m_wf[static_cast<int> ( 'B' ) ] = 2.12;
   m_wf[static_cast<int> ( 'Y' ) ] = 2.00;
   m_wf[static_cast<int> ( 'F' ) ] = 1.47;
   m_wf[static_cast<int> ( 'V' ) ] = 1.07;
   m_wf[static_cast<int> ( 'W' ) ] = 0.94;
   m_wf[static_cast<int> ( 'K' ) ] = 0.84;
   m_wf[static_cast<int> ( 'X' ) ] = 0.35;
   m_wf[static_cast<int> ( 'Z' ) ] = 0.24;
   m_wf[static_cast<int> ( 'Q' ) ] = 0.23;
   m_wf[static_cast<int> ( 'J' ) ] = 0.15;

   for ( int i=0; i<256; ++i )
   {
      m_wr[i] = 0.8 * m_wf[i];
   }

}//===============================================================================================
void m_fnSetVowelsDown( )
{

   int i;
   for ( i=0; i<256; ++i )
   {
      m_wf[i] = 1.0;
   }
   m_wf[static_cast<int> ( 'e' ) ] = 0.5;
   m_wf[static_cast<int> ( 'a' ) ] = 0.5;
   m_wf[static_cast<int> ( 'i' ) ] = 0.5;
   m_wf[static_cast<int> ( 'o' ) ] = 0.5;
   m_wf[static_cast<int> ( 'u' ) ] = 0.5;
   m_wf[static_cast<int> ( 'y' ) ] = 0.5;

   m_wf[static_cast<int> ( 'E' ) ] = 0.5;
   m_wf[static_cast<int> ( 'A' ) ] = 0.5;
   m_wf[static_cast<int> ( 'I' ) ] = 0.5;
   m_wf[static_cast<int> ( 'O' ) ] = 0.5;
   m_wf[static_cast<int> ( 'U' ) ] = 0.5;
   m_wf[static_cast<int> ( 'Y' ) ] = 0.5;

   for ( i=0; i<256; ++i )
   {
      m_wr[i] = 0.8 * m_wf[i];
   }

}

//===============================================================================================
private:
   valarray<double> m_wf ;
   valarray<double> m_wr ;
   valarray<double> m_wc ;
   size_t m_lMaxCharacter;
   bool m_bDoCompensation;

//===============================================================================================
   // compensation allows some extra credit to be added for missing characters. The problem
   // with compensation is that the triangle rule will no longer apply if compensation is used.
   // If your needs are this strong, you should probably go to the use of Levenstein edit 
   // distance instead of this simple string distance algorithm    lca  12/5/2001
//===============================================================================================
   double compensation (
                     const size_t& lLength_s,
                     const size_t& lLength_t,
                     const valarray<size_t>& cnts,
                     const valarray<size_t>& cntt ) const
   {
      double sum = 0.0;
      size_t i;
      for ( i=0; i<m_lMaxCharacter;  ++i )
      {
         if ( cnts[i] == cntt[i] )
         {
            // nothing to do
         }
         else if ( cnts[i] > cntt[i] )
         {
            sum += lLength_s * (  cnts[i] - cntt[i] );
         }
         else
         {
            sum += lLength_t * (  cntt[i] - cnts[i] );
         }
      }
      return ( sum );
   }

//===============================================================================================
   double totalM ( const T& x ) const
   {
      size_t lParticularCharacter;
      double dReturn = 0;
      const size_t lLengthOfString = x.length( );

      for ( size_t i=0; i<lLengthOfString; ++i )
      {
         lParticularCharacter = static_cast<long> (x[i] );
         dReturn += ( lLengthOfString - i ) * m_wf[ lParticularCharacter ] + 
                                  ( i + 1 ) * m_wr[ lParticularCharacter ];
      }
      return ( dReturn );
   };

//===============================================================================================
   double m( const T& s, const T& t, const int& direction, const valarray<double>& w ) const
   {  // this code could probably get into trouble if characters are signed char
       double sum = 0.0;
       size_t i;
       const long lIncrement = ( direction>0 ? 1 : -1 );
    
       valarray<size_t> cnts ( m_lMaxCharacter );
       valarray<size_t> cntt ( m_lMaxCharacter );
       cnts = 0;
       cntt = 0;
    
       long lpts = ( lIncrement < 1 ? s.length( ) : -1 );
       long lptt = ( lIncrement < 1 ? t.length( ) : -1 );
       long s_count = s.length( ) + 1;
       long t_count = t.length( ) + 1;
       size_t s_char;
       size_t t_char;

       //cout << endl << endl;
       //cout << " increment " << direction << endl;
       //cout << " s \"" << s << "\"  t \"" << t << "\"" << endl;
       //cout << " i \t lpts \t lptt \t s_count  t_count " << endl;
       for ( i=0; i<max( s.length( ), t.length( ) ); i++ )
       {
           lpts += lIncrement;
           lptt += lIncrement;
           s_count = max( s_count-1, 0 );
           t_count = max( t_count-1, 0 );
           //cout << i << " \t" << lpts << " \t" << lptt << " \t" << s_count << " \t" << t_count << endl;
           if ( s_count > 0 )
           {
               s_char = static_cast<size_t> ( s[lpts] );
               cnts[s_char] += 1;
           }
           if ( t_count > 0)
           {
               t_char = static_cast<size_t> ( t[lptt] );
               cntt[t_char] += 1;
           }
           if ( ( s_count > 0 ) && ( t_count > 0 ) && ( s_char == t_char ))
               sum += w[s_char] * ( s_count+t_count );
           else
           {
               if ( ( s_count > 0 ) && ( cnts[s_char] <= cntt[s_char] )) 
                   sum += w[s_char] * ( s_count+t_count );
               if ( ( t_count > 0 ) && ( cntt[t_char] <= cnts[t_char] )) 
                   sum += w[t_char] * ( s_count+t_count ) ;
           }
       }
       if ( m_bDoCompensation && direction > 0 )  //  only do compensation once (if at all) !
          sum += + compensation( lpts, lptt, cnts, cntt );
       return ( sum );
   };

}; // template class CThetaMatch

/*

Here's a list of the DOS command names:
APPEND ASSIGN ATTRIB BACKUP BREAK BUFFERS CALL CHCP CHDIR CHKDSK CHOICE
CLS COMMAND COMP COPY COUNTRY CTTY DATE DBLSPACE DEBUG DEFRAG DEL
ERASE DELOLDOS DELTREE DEVICE DEVICEHIGH DIR DISKCOMP DISKCOPY DOS
DOSKEY DOSSHELL DRIVPARM ECHO EDIT EMM386 EXE2BIN EXIT EXPAND FASTHELP
FASTOPEN FC FCBS FDISK FILES FIND FOR FORMAT GOTO GRAFTABL GRAPHICS HELP
IF INCLUDE INSTALL INTERLINK INTERSVR JOIN KEYB LABEL LASTDRIVE LOADFIX
LOADHIGH MEM MEMMAKER MENUCOLOR MENUDEFAULT MENUITEM MIRROR MKDIR
MODE MORE MOVE MSAV MSBACKUP MSCDEX MSD NLSFUNC NUMLOCK PATH PAUSE
POWER PRINT PROMPT RECOVER REM RENAME REN REPLACE RESTORE RMDIR RD
SCANDISK SELECT SET SETVER SHARE SHELL SHIFT SORT STACKS SUBMENU SUBST
SWITCHES SYS TIME TREE TYPE UNDELETE UNFORMAT VER VERIFY VOL VSAFE
XCOPY

*/
#endif // THETA_H_INCLUDED

