//
// TimeShareGASolver.cpp : CTimeShareGASolver
////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "tsgenome.h"
#include "gasolver.h"
#include "Tsgasolve.h"
#include "TimeShareGASolver.h"


#import "msado15.dll" rename("EOF", "adoEOF" ) no_namespace
#include "assert.h"

/////////////////////////////////////////////
STDMETHODIMP CTimeShareGASolver::Solve()
{
   // Read the data.
   ReadOwnerInfo();
   //
   // Set the parameters.
   GASolver <TSGenome> gaSolver;
   TSGenome bestSoln;
   gaSolver.SetPopSize(200);
   gaSolver.SetNumGenerations(2000);
   gaSolver.SetMutationFactor(0.01);
   gaSolver.SetCrossOverFactor(0.30);
   gaSolver.SetFitnessSigma(3.0);
   //
   // Solve for minimum.
   gaSolver.Minimize(bestSoln);
   //
   // Update database.
   WriteBestSolution(bestSoln);
   return S_OK;
}


/////////////////////////////////////////////
HRESULT CTimeShareGASolver::ReadOwnerInfo()
{
   //
   // Read the database.
   try
   {
      _RecordsetPtr pRs;
      pRs.CreateInstance(__uuidof(Recordset));
      //
      // Create the query.
      CComBSTR bstrQuery = L"select OwnerID, Occupancy, FirstChoice, SecondChoice, ThirdChoice from Owners order by Occupancy DESC, OwnerID";
      _variant_t vBuff= bstrQuery;
       //
       // Open the recordset;
      pRs->Open(vBuff, L"dsn=TSDATA;pwd=;",adOpenStatic, adLockOptimistic, adCmdUnknown);
       //
       // Get the row count;
      _variant_t cnt = pRs->GetRecordCount();
      TSGenome genome;
       //
       // Get the data.
      _variant_t vArr = pRs->GetRows( cnt);
      //
      //  Pull data out of array.
      long lFields, lRecords;
      VARIANT vRet;
      HRESULT hr = SafeArrayGetUBound(vArr.parray, 1, &lFields);
      hr = SafeArrayGetUBound(vArr.parray, 2, &lRecords);
      long ind[2];
      for(int i=0;i<=lRecords;i++)   //-- Next Row loop
      { 
         for(int j=0;j<=lFields;j++)   //-- Next Column loop
         {
            ind[0]=j; ind[1]=i;
            hr = SafeArrayGetElement(vArr.parray, ind, &vRet);
            //-- here you can use returned VARIANT vRet
            switch (j)
            {
               case 0:
               genome.m_owner[i].m_ownerID         = vRet.lVal;
               break;

               case 1:
               genome.m_owner[i].m_occupCnt   = vRet.lVal;
               break;

               case 2:
               genome.m_owner[i].m_choice[0]     = vRet.lVal-1;
               break;

               case 3:
               genome.m_owner[i].m_choice[1]     = vRet.lVal-1;
               break;

               case 4:
               genome.m_owner[i].m_choice[2]     = vRet.lVal-1;
               break;
            }
         }
      }
   }
   catch ( _com_error e )
   {
      return e.Error();
   }
   catch ( ... )
   {
      return E_FAIL;
   }
   return S_OK;
}


/////////////////////////////////////////////
CTimeShareGASolver::WriteBestSolution(TSGenome best)
{
   //
   // Update the database.
   try
   {
      _RecordsetPtr pRs;
      pRs.CreateInstance(__uuidof(Recordset));
      //
      // Create the query.
      CComBSTR bstrQuery = L"select OwnerID, Period, Cost from Owners";
      _variant_t vBuff = bstrQuery;
       //
       // Open the recordset;
      pRs->Open(vBuff, L"dsn=TSDATA;pwd=;",adOpenStatic, adLockOptimistic, adCmdUnknown);
       //
       // Get the row count;
      _variant_t cnt = pRs->GetRecordCount();
      TSGenome genome;
      while (VARIANT_FALSE == pRs->adoEOF)
      {
         _variant_t itemNo = (long)0;
         int ownerID = pRs->Fields->GetItem(itemNo)->Value.lVal;
         int period=0;
         int cost=0;
         for (int i=0;i<MAX_OWNERS;i++)
         {
            if (best.m_owner[i].m_ownerID == ownerID)
            {
               period = best.m_assign[i];
               cost = best.CalcOwnerCost(i, period);
               break;
            }
         }
         //
         //  Update the period
         itemNo = (long)1;
         _variant_t newVal = (long)(period + 1);
         pRs->Fields->GetItem(itemNo)->PutValue( newVal);
         //
         //  Update the cost.
         itemNo = (long)2;
         newVal = (long)cost;
         pRs->Fields->GetItem(itemNo)->PutValue( newVal);
         pRs->Update();
         pRs->MoveNext();   
      }
   }
   catch ( _com_error e )
   {
      return e.Error();
   }
   catch ( ... )
   {
      return E_FAIL;
   }
   return S_OK;
}

