//
//  tsgenome.cpp - time share assignment
//  genome represents solution to 
//  time share assignment problem.
////////////////////////////////////////
#include "stdafx.h"
#include "tsgenome.h"
#include "assert.h"
Owner    TSGenome::m_owner[MAX_OWNERS];

//-------------------------------------
TSGenome::TSGenome()
{
   UnassignAll();
}

//-------------------------------------
TSGenome& TSGenome::operator = (const TSGenome& rhs)
{
   m_obj = rhs.m_obj;
   m_fitness = rhs.m_fitness;
   m_selProb = rhs.m_selProb;
   m_selected = rhs.m_selected;
   memcpy( m_assign, rhs.m_assign, MAX_OWNERS*sizeof(int));
   memcpy( m_period, rhs.m_period, MAX_PERIODS*sizeof(Period)) ;
   m_assnTotal = rhs.m_assnTotal;
   return *this;
}

//-------------------------------------
void TSGenome::UnassignAll()
{
   for (int i=0;i<MAX_OWNERS;i++)
      m_assign[i] = -1;
   for (int j=0;j<MAX_PERIODS;j++)
   {
      for (int k=0;k<MAX_UNITS;k++)
      {
         m_period[j].m_owner[k] = -1;
      }
      m_period[j].m_assnCnt = 0;
      m_period[j].m_assnTotal = 0;
   }
   m_assnTotal=0;
}

//-------------------------------------
void TSGenome::Assign(int owner, int period)
{
   assert(m_period[period].m_assnCnt < MAX_UNITS);
   assert(m_assign[owner] == -1);
   m_assign[owner] = period;
   for (int k=0;k<MAX_UNITS;k++)
   {
      if (m_period[period].m_owner[k] == -1)
         break;
   }
   assert( k <MAX_UNITS);
   m_period[period].m_owner[k] = owner;
   m_period[period].m_assnCnt++;
   m_period[period].m_assnTotal += m_owner[owner].m_occupCnt;
   m_assnTotal++;
}

//-------------------------------------
void TSGenome::Unassign(int owner)
{
   int period = m_assign[owner];
   assert(period >= 0);
   assert(period < MAX_PERIODS);
   m_assign[owner] = -1;
   for (int k=0;k<MAX_UNITS;k++)
   {
      if (m_period[period].m_owner[k] == owner)
         break;
   }
   assert( k <MAX_UNITS);
   m_period[period].m_owner[k] = -1;
   m_period[period].m_assnCnt--;
   m_period[period].m_assnTotal -= m_owner[owner].m_occupCnt;
   m_assnTotal--;
}


//-------------------------------------
void   TSGenome::Initialize()
{
   UnassignAll();
   while (m_assnTotal < MAX_OWNERS)
   {
      //  Do greedy assignment at random.
      int owner = RandomNum(MAX_OWNERS);
      if (m_assign[owner] == -1)
      {
         DoGreedyAssign( owner);
      }
   }
}

//-------------------------------------
void TSGenome::DoGreedyAssign(int i)
{
   assert(m_assign[i] == -1);
   Owner* owner = &m_owner[i];
   int occup = owner->m_occupCnt;
   int period = owner->m_choice[0];
   if ((m_period[period].m_assnCnt == MAX_UNITS) 
   || ((m_period[period].m_assnTotal + occup) > MAX_OCCUPANTS))
   {
      period = owner->m_choice[1];
      if ((m_period[period].m_assnCnt == MAX_UNITS)
      || ((m_period[period].m_assnTotal + occup) > MAX_OCCUPANTS))
      {
         period = owner->m_choice[2];
         if ((m_period[period].m_assnCnt == MAX_UNITS)
         || ((m_period[period].m_assnTotal + occup) > MAX_OCCUPANTS))
         {

            //
            // Couldn't find a preferred slot.
            // Put him in least occupied slot.
            int bestPeriod = -1;
            int leastOccup = 1000000;
            for (period =0;period < MAX_PERIODS;period++)
            {
               if ((m_period[period].m_assnCnt < MAX_UNITS)
                && (m_period[period].m_assnTotal < leastOccup))
               {
                  bestPeriod = period;
                  leastOccup = m_period[period].m_assnTotal;
               }
            }
            period = bestPeriod;
         }
      }
   }
   Assign(i, period);
}

//-------------------------------------
void   TSGenome::Mutate( double mutationFactor)
{
   //
   double pk = RandomNum(10000)/10000.0;
   if (pk < mutationFactor)
   {
      //unassign 20 owners in random order
      int cnt =0;
      while (m_assnTotal > MAX_OWNERS - 20)
      {
         int owner = RandomNum(MAX_OWNERS);
         if (m_assign[owner] != -1)
            Unassign(owner);
      }
      while (m_assnTotal < MAX_OWNERS)
      {
         //
         //  Now do greedy assignment of these 
         //  owners in random order.
         int owner = RandomNum(MAX_OWNERS);
         if (m_assign[owner] == -1)
         {
            DoGreedyAssign( owner);
         }
      }
   }
}

//-------------------------------------
void TSGenome::CrossOver( GAGenome *genome)
{
   TSGenome *child;
   child = (TSGenome *)genome;
   int period1 = RandomNum(MAX_PERIODS);
   for (int unit =0;unit<MAX_UNITS;unit++)
   {
      //  For each period1 owner in
      //  this genome.
      int owner = m_period[period1].m_owner[unit];
      //
      //  If he's not period1 in the
      //  child already then  unassign him
      //  so he can be reassigned to period1
      if (child->m_assign[owner] != period1)
      {
         child->Unassign(owner);
      }
   }
   for (unit =0;unit<MAX_UNITS;unit++)
   {
      //  For each period1 owner in child
      int owner = child->m_period[period1].m_owner[unit];
      //
      //  If he's not period1 in parent then
      //  unassign him in child to make 
      //  room in period1.
      if (m_assign[owner] != period1)
      {
         child->Unassign(owner);
      }
   }
   for (unit =0;unit<MAX_UNITS;unit++)
   {
      //  Assign all the period1 owners.
      int owner = m_period[period1].m_owner[unit];
      if (child->m_assign[owner] != period1)
      {
         child->Assign(owner, period1);
      }
   }
   //
   //  Reassign the remaining unassigned. 
   //  Greedily but in random order.
   while (child->m_assnTotal < MAX_OWNERS)
   {
      int owner = RandomNum(MAX_OWNERS);
      if (child->m_assign[owner] == -1)
      {
         child->DoGreedyAssign( owner);
      }
   }
}


//-------------------------------------
double   TSGenome::Evaluate()
{
   //
   // calc the assigment penalty for each owner.
   double totalCost = 0;
   for (int i=0;i<MAX_OWNERS;i++)
   {
      int assign = m_assign[i];
      totalCost += CalcOwnerCost( i, assign);
   }
   //
   //  Have the assigment cost. Now calc penalty cost.
   for (int p=0;p<MAX_PERIODS;p++)
   {
      int overOcc = m_period[p].m_assnTotal 
                  - MAX_OCCUPANTS;
       // penalty for too many occupants.
      if (overOcc > 0)
         totalCost += 40;
   }
   return totalCost;
}

//-------------------------------------
int TSGenome::CalcOwnerCost(int ownr, int assign)
{
   Owner *owner = &m_owner[ownr];
   int cost;
   if (assign == owner->m_choice[0])
      cost = 0;
   else if (assign == owner->m_choice[1])
      cost = 2*owner->m_occupCnt;
   else if (assign == owner->m_choice[2])
      cost = 4*owner->m_occupCnt;
   else
      cost = 50 + 7*owner->m_occupCnt;
   return cost;
}

